package org.anarres.iengine.preprocess;

import java.io.IOException;
import java.util.Stack;
import static org.anarres.iengine.preprocess.Token.*;

%%

%class Lexer
%unicode
%line
%column
%final
%buffer 1024
%type Token
%debug

%yylexthrow{
LexerException
%yylexthrow}

%{
	private StringBuilder	buf = new StringBuilder();

	private String string() {
		String	str = buf.toString();
		buf.setLength(0);
		return str;
	}

	private Token token(int type) {
		return new Token(type, yyline, yycolumn);
	}

	private Token token(int type, Object value) {
		return new Token(type, yyline, yycolumn, value);
	}

	private void exception(String msg) throws LexerException {
		throw new LexerException(yyline, yycolumn, msg);
	}

	public void expect(int type)
						throws IOException,
								LexerException {
		Token	token = yylex();
		if (token.getType() != type)
			exception("Expected token type " + type +
							" but got " + token);
	}

	public Token yylexsw()
						throws IOException,
								LexerException {
		Token	token = yylex();
		while (token.getType() == WHITESPACE)
			token = yylex();
		return token;
	}
%}

/* main character classes */
LineTerminator = \r|\n|\r\n|\u2028|\u2029|\u000B|\u000C|\u0085
InputCharacter = [^\r\n]

WhiteSpace = {LineTerminator} | [ \t\f]

/* comments */
Comment = {TraditionalComment} | {EndOfLineComment}

TraditionalComment = "/*" ~"*/"
EndOfLineComment = "//" {InputCharacter}* {LineTerminator}?

/* identifiers */
Identifier = [:jletter:][:jletterdigit:]*

/* integer literals */
DecIntegerLiteral = 0 | [1-9][0-9]*
DecLongLiteral    = {DecIntegerLiteral} [lL]

HexIntegerLiteral = 0 [xX] 0* {HexDigit} {1,8}
HexLongLiteral    = 0 [xX] 0* {HexDigit} {1,16} [lL]
HexDigit          = [0-9a-fA-F]

OctIntegerLiteral = 0+ [1-3]? {OctDigit} {1,15}
OctLongLiteral    = 0+ 1? {OctDigit} {1,21} [lL]
OctDigit          = [0-7]

/* string and character literals */
StringCharacter = [^\r\n\"\\]
SingleCharacter = [^\r\n\'\\]

%s STRING, CHAR, PP

%%

\\{LineTerminator}   { }

<YYINITIAL> {
	{Comment}            { return token(COMMENT, yytext()); }
	{WhiteSpace}+        { return token(WHITESPACE, yytext()); }
	{Identifier}         { return token(IDENTIFIER, yytext()); }

	^[:space:]*#         { return token(HASH); }

	.                    { return token(LITERAL, yytext()); }
}

<YYINITIAL,PP> {
	\"                   { yybegin(STRING); }
	\'                   { yybegin(CHAR); }
}

<PP> {
	{LineTerminator}     { yybegin(YYINITIAL); return token(WHITESPACE, yytext()); }

	{DecIntegerLiteral}  { return token(NUMBER, Long.parseLong(yytext())); }
	{DecLongLiteral}     { return token(LITERAL, yytext()); }
	{HexIntegerLiteral}  { return token(NUMBER, Long.parseLong(yytext().substring(2), 16)); }
	{HexLongLiteral}     { return token(LITERAL, yytext()); }
	{OctIntegerLiteral}  { return token(NUMBER, Long.parseLong(yytext(), 8)); }
	{OctLongLiteral}     { return token(LITERAL, yytext()); }

	/* We only have to recognise things that the preprocessor
	 * cares about. For all other purposes, we can just let the
	 * characters go through unrecognised. */
	##                   { return token(PASTE); }
	&&                   { return token(AND); }
	\|\|                 { return token(OR); }
	==                   { return token(EQ); }
	\!=                  { return token(NE); }
	>=                   { return token(GE); }
	\<=                  { return token(LE); }
	[<>\.=+-/%*&|~\^]    { return token(yycharat(0)); }
	[\[\]{}()]           { return token(yycharat(0)); }
	[!?:;]               { return token(yycharat(0)); }
}

<STRING> {
	\"                   {
		yybegin(YYINITIAL);
		return token(STRING, string());
	}

	{StringCharacter}+   { buf.append( yytext() ); }

	/* escape sequences */
	"\\b"                { buf.append( '\b' ); }
	"\\t"                { buf.append( '\t' ); }
	"\\n"                { buf.append( '\n' ); }
	"\\f"                { buf.append( '\f' ); }
	"\\r"                { buf.append( '\r' ); }
	"\\\""               { buf.append( '\"' ); }
	"\\'"                { buf.append( '\'' ); }
	"\\\\"               { buf.append( '\\' ); }
	\\[0-3]?{OctDigit}?{OctDigit}  {
		char val = (char)Integer.parseInt(yytext().substring(1), 8);
		buf.append( val );
	}

	/* error cases */
	\\.                            {
		exception("Illegal escape sequence \""+yytext()+"\"");
	}
	{LineTerminator}               {
		 exception("Unterminated string at end of line");
	}
}

<CHAR> {

	{SingleCharacter}\'            { yybegin(YYINITIAL); return token(CHARACTER, new Character(yytext().charAt(0))); }

	/* escape sequences */
	"\\b"\'                        { yybegin(YYINITIAL); return token(CHARACTER, new Character('\b'));}
	"\\t"\'                        { yybegin(YYINITIAL); return token(CHARACTER, new Character('\t'));}
	"\\n"\'                        { yybegin(YYINITIAL); return token(CHARACTER, new Character('\n'));}
	"\\f"\'                        { yybegin(YYINITIAL); return token(CHARACTER, new Character('\f'));}
	"\\r"\'                        { yybegin(YYINITIAL); return token(CHARACTER, new Character('\r'));}
	"\\\""\'                       { yybegin(YYINITIAL); return token(CHARACTER, new Character('\"'));}
	"\\'"\'                        { yybegin(YYINITIAL); return token(CHARACTER, new Character('\''));}
	"\\\\"\'                       { yybegin(YYINITIAL); return token(CHARACTER, new Character('\\')); }
	\\[0-3]?{OctDigit}?{OctDigit}\' {
		yybegin(YYINITIAL); 
		int val = Integer.parseInt(yytext().substring(1,yylength()-1),8);
		return token(CHARACTER, new Character((char)val));
	}

	/* error cases */
	\\.                            { exception("Illegal escape sequence \""+yytext()+"\""); }
	{LineTerminator}               { exception("Unterminated character literal at end of line"); }
}

.|\n         { exception("Illegal character \""+yytext()+ "\""); }
<<EOF>>      { return token(EOF); }
